import os
import re
import subprocess
import typing as t
from unittest import mock

import pytest
from pytest import param

import shelmet as sh


parametrize = pytest.mark.parametrize


@pytest.fixture()
def mock_subprocess_run() -> t.Generator[mock.MagicMock, None, None]:
    with mock.patch("subprocess.run") as _mock_subprocess_run:
        yield _mock_subprocess_run


def run_call_args(
    args,
    stdin=None,
    input=None,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
    cwd=None,
    timeout=None,
    check=True,
    encoding=None,
    errors=None,
    text=True,
    env=None,
    **popen_kwargs,
):
    popen_kwargs.pop("capture_output", None)
    popen_kwargs.pop("combine_output", None)
    popen_kwargs.pop("replace_env", None)
    return mock.call(
        args,
        stdin=stdin,
        input=input,
        stdout=stdout,
        stderr=stderr,
        cwd=cwd,
        timeout=timeout,
        check=check,
        encoding=encoding,
        errors=errors,
        universal_newlines=text,
        env=env,
        **popen_kwargs,
    )


def test_command__returns_command_object_with_defaults():
    args = ["ls", "-la"]
    cmd = sh.cmd(*args)

    assert cmd.args == args
    assert cmd.stdin is None
    assert cmd.input is None
    assert cmd.stdout == subprocess.PIPE
    assert cmd.stderr == subprocess.PIPE
    assert cmd.capture_output is True
    assert cmd.cwd is None
    assert cmd.timeout is None
    assert cmd.check is True
    assert cmd.encoding is None
    assert cmd.errors is None
    assert cmd.text is True
    assert cmd.env is None
    assert cmd.replace_env is False
    assert cmd.popen_kwargs == {}
    assert cmd.parent is None
    assert cmd.parents == []
    assert cmd.shell_cmd == "ls -la"


@parametrize(
    "args, exception, match",
    [
        param([], TypeError, "Command(): requires at least one non-empty positional argument"),
        param(
            [None, None],
            TypeError,
            "Command(): requires at least one non-empty positional argument",
        ),
        param(
            [5],
            TypeError,
            "Command(): requires all positional arguments to be either string or bytes",
        ),
    ],
)
def test_command__raises_on_bad_args(args: list, exception: t.Type[Exception], match: str):
    with pytest.raises(exception, match=re.escape(match)):
        sh.cmd(*args)


@parametrize(
    "cmd, expected_repr",
    [
        param(sh.cmd("ls", "-la"), "Command(args=['ls', '-la'])"),
        param(
            sh.cmd("ls", "-la").pipe("grep", "foo"),
            "Command(args=['grep', 'foo'], parents=[PipeCommand(args=['ls', '-la'])])",
        ),
        param(
            sh.cmd("cmd1").pipe("cmd2").pipe("cmd3"),
            (
                "Command(args=['cmd3'],"
                " parents=[PipeCommand(args=['cmd1']), PipeCommand(args=['cmd2'])])"
            ),
        ),
    ],
)
def test_command_repr(cmd, expected_repr):
    assert repr(cmd) == expected_repr


@parametrize(
    "args, kwargs, expected_call",
    [
        param(["ls"], {}, run_call_args(["ls"]), id="single_arg"),
        param(["ls", "-l", "-a"], {}, run_call_args(["ls", "-l", "-a"]), id="multiple_args"),
        param(
            [["ls", "-l", "-a"]], {}, run_call_args(["ls", "-l", "-a"]), id="single_list_of_args"
        ),
        param(
            [["ls"], ["-l", "-a"]],
            {},
            run_call_args(["ls", "-l", "-a"]),
            id="multiple_lists_of_args",
        ),
        param(
            ["ls", None, "-l", None, "-a"],
            {},
            run_call_args(["ls", "-l", "-a"]),
            id="none_valued_args_discarded",
        ),
        param(
            ["ls"],
            {"capture_output": False},
            run_call_args(["ls"], stdout=None, stderr=None),
            id="no_capture_output",
        ),
        param(["ls"], {"stdout": None}, run_call_args(["ls"], stdout=None), id="no_capture_stdout"),
        param(["ls"], {"stderr": None}, run_call_args(["ls"], stderr=None), id="no_capture_stderr"),
        param(
            ["ls"],
            {"combine_output": True},
            run_call_args(["ls"], stderr=subprocess.STDOUT),
            id="combine_output",
        ),
        param(["ls"], {"text": False}, run_call_args(["ls"], text=False), id="no_text"),
        param(["ls"], {"input": "test"}, run_call_args(["ls"], input="test"), id="input_as_str"),
        param(
            ["ls"],
            {"input": b"test", "text": False},
            run_call_args(["ls"], input=b"test", text=False),
            id="input_as_bytes",
        ),
        param(
            ["ls"],
            {"input": b"test"},
            run_call_args(["ls"], input="test"),
            id="coerce_input_to_str",
        ),
        param(
            ["ls"],
            {"input": "test", "text": False},
            run_call_args(["ls"], input=b"test", text=False),
            id="coerce_input_to_bytes",
        ),
        param(
            ["ls"],
            {"stdin": subprocess.DEVNULL},
            run_call_args(["ls"], stdin=subprocess.DEVNULL),
            id="set_stdin",
        ),
        param(["ls"], {"cwd": "."}, run_call_args(["ls"], cwd="."), id="set_cwd"),
        param(["ls"], {"timeout": 10}, run_call_args(["ls"], timeout=10), id="set_timeout"),
        param(
            ["ls"],
            {"encoding": "utf-8"},
            run_call_args(["ls"], encoding="utf-8"),
            id="set_encoding",
        ),
        param(
            ["ls"],
            {"errors": "strict"},
            run_call_args(["ls"], errors="strict"),
            id="set_errors",
        ),
    ],
)
def test_command_run__passes_arguments_to_subprocess_run(
    mock_subprocess_run: mock.MagicMock, args: list, kwargs: dict, expected_call: tuple
):
    sh.run(*args, **kwargs)
    assert mock_subprocess_run.call_args == expected_call


def test_command_run__extends_env(mock_subprocess_run: mock.MagicMock):
    env = {"a": "1", "b": "2"}
    expected_call = run_call_args(["ls"], env={**os.environ, **env})
    sh.run("ls", env=env)
    assert mock_subprocess_run.call_args == expected_call


def test_command_run__replaces_env(mock_subprocess_run: mock.MagicMock):
    env = {"a": "1", "b": "2"}
    expected_call = run_call_args(["ls"], env=env)
    sh.run("ls", env=env, replace_env=True)
    assert mock_subprocess_run.call_args == expected_call


@parametrize(
    "cmd, extra_args, overrides, expected_call",
    [
        param(sh.cmd("ls"), ["-l", "-a"], {}, run_call_args(["ls", "-l", "-a"])),
        param(
            sh.cmd("ls"),
            [],
            {"env": {"A": "B"}, "replace_env": True},
            run_call_args(["ls"], env={"A": "B"}),
        ),
        param(sh.cmd("ls"), [], {"cwd": "/tmp"}, run_call_args(["ls"], cwd="/tmp")),
        param(
            sh.cmd("ls"),
            [],
            {"stdin": subprocess.PIPE, "stdout": None, "stderr": None},
            run_call_args(["ls"], stdin=subprocess.PIPE, stdout=None, stderr=None),
        ),
        param(
            sh.cmd("ls"),
            [],
            {"capture_output": False},
            run_call_args(["ls"], stdout=None, stderr=None),
        ),
        param(
            sh.cmd("ls"),
            [],
            {"input": "test", "timeout": 10},
            run_call_args(["ls"], input="test", timeout=10),
        ),
        param(
            sh.cmd("ls"),
            [],
            {"encoding": "utf-8", "errors": "ignore", "text": False},
            run_call_args(["ls"], encoding="utf-8", errors="ignore", text=False),
        ),
    ],
)
def test_command_run__overrides_defaults(
    mock_subprocess_run: mock.MagicMock,
    cmd: sh.Command,
    extra_args: list,
    overrides: dict,
    expected_call: tuple,
):
    cmd.run(*extra_args, **overrides)
    assert mock_subprocess_run.called
    assert mock_subprocess_run.call_args == expected_call


@parametrize(
    "cmd, mock_side_effect, expected_call_args_list",
    [
        param(
            sh.cmd("cmd1").pipe("cmd2").pipe("cmd3"),
            [
                subprocess.CompletedProcess(args=["cmd1"], returncode=0),
                subprocess.CompletedProcess(args=["cmd2"], returncode=0),
                subprocess.CompletedProcess(args=["cmd3"], returncode=0),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="pipe_run_3",
        ),
        param(
            sh.cmd("cmd1").pipe("cmd2").pipe("cmd3"),
            [
                subprocess.CalledProcessError(cmd=["cmd1"], returncode=1),
                subprocess.CalledProcessError(cmd=["cmd2"], returncode=1),
                subprocess.CompletedProcess(args=["cmd3"], returncode=0),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="pipe_run_3",
        ),
        param(
            sh.cmd("cmd1").and_("cmd2").and_("cmd3"),
            [subprocess.CompletedProcess(args=["cmd1"], returncode=1)],
            [["cmd1"]],
            id="and_run_1",
        ),
        param(
            sh.cmd("cmd1").and_("cmd2").and_("cmd3"),
            [
                subprocess.CompletedProcess(args=["cmd1"], returncode=0),
                subprocess.CompletedProcess(args=["cmd2"], returncode=1),
            ],
            [["cmd1"], ["cmd2"]],
            id="and_run_2",
        ),
        param(
            sh.cmd("cmd1").and_("cmd2").and_("cmd3"),
            [
                subprocess.CompletedProcess(args=["cmd1"], returncode=0),
                subprocess.CompletedProcess(args=["cmd2"], returncode=0),
                subprocess.CompletedProcess(args=["cmd3"], returncode=0),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="and_run_3",
        ),
        param(
            sh.cmd("cmd1").or_("cmd2").or_("cmd3"),
            [subprocess.CompletedProcess(args=["cmd1"], returncode=0)],
            [["cmd1"]],
            id="or_run_1",
        ),
        param(
            sh.cmd("cmd1").or_("cmd2").or_("cmd3"),
            [
                subprocess.CompletedProcess(args=["cmd1"], returncode=1),
                subprocess.CompletedProcess(args=["cmd2"], returncode=0),
            ],
            [["cmd1"], ["cmd2"]],
            id="or_run_2",
        ),
        param(
            sh.cmd("cmd1").or_("cmd2").or_("cmd3"),
            [
                subprocess.CompletedProcess(args=["cmd1"], returncode=1),
                subprocess.CompletedProcess(args=["cmd2"], returncode=1),
                subprocess.CompletedProcess(args=["cmd3"], returncode=0),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="or_run_3",
        ),
        param(
            sh.cmd("cmd1").or_("cmd2").or_("cmd3"),
            [
                subprocess.CalledProcessError(cmd=["cmd1"], returncode=1),
                subprocess.CalledProcessError(cmd=["cmd2"], returncode=1),
                subprocess.CompletedProcess(args=["cmd3"], returncode=1),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="or_run_3_errors",
        ),
        param(
            sh.cmd("cmd1").after("cmd2").after("cmd3"),
            [
                subprocess.CompletedProcess(args=["cmd1"], returncode=1),
                subprocess.CompletedProcess(args=["cmd2"], returncode=1),
                subprocess.CompletedProcess(args=["cmd3"], returncode=1),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="after_run_3_failures",
        ),
        param(
            sh.cmd("cmd1").after("cmd2").after("cmd3"),
            [
                subprocess.CalledProcessError(cmd=["cmd1"], returncode=1),
                subprocess.CalledProcessError(cmd=["cmd2"], returncode=1),
                subprocess.CompletedProcess(args=["cmd3"], returncode=1),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="after_run_3_errors",
        ),
        param(
            sh.cmd("cmd1").after("cmd2").after("cmd3"),
            [
                subprocess.CompletedProcess(args=["cmd1"], returncode=0),
                subprocess.CompletedProcess(args=["cmd2"], returncode=0),
                subprocess.CompletedProcess(args=["cmd3"], returncode=0),
            ],
            [["cmd1"], ["cmd2"], ["cmd3"]],
            id="after_run_3_successes",
        ),
    ],
)
def test_command_run__calls_parent_command_run(
    mock_subprocess_run: mock.MagicMock,
    cmd: sh.Command,
    mock_side_effect: list,
    expected_call_args_list: list,
):
    if mock_side_effect:
        mock_subprocess_run.side_effect = mock_side_effect

    result = cmd.run()
    assert len(mock_subprocess_run.call_args_list) == len(expected_call_args_list)
    assert result.args == expected_call_args_list[-1]

    for i, call_args in enumerate(expected_call_args_list):
        call_cmd = mock_subprocess_run.call_args_list[i]
        assert call_args == call_cmd[0][0]


def test_command_run__pipes_parent_stdout_to_child(mock_subprocess_run: mock.MagicMock):
    cmd1_stdout = "cmd1_stdout"
    mock_subprocess_run().stdout = cmd1_stdout

    cmd = sh.cmd("cmd1").pipe("cmd2")
    cmd.run()

    call_cmd2 = mock_subprocess_run.call_args_list[-1]
    assert call_cmd2[0][0] == ["cmd2"]
    assert call_cmd2[1]["input"] == cmd1_stdout


def test_command_pipe__sets_parent():
    cmd1 = sh.cmd("cmd1")
    cmd2 = cmd1.pipe("cmd2")
    cmd3 = cmd2.pipe("cmd3")
    cmd4 = cmd3.pipe("cmd4")

    assert cmd2.parent.command is cmd1
    assert cmd3.parent.command is cmd2
    assert cmd4.parent.command is cmd3


def test_command_pipe__returns_child_command():
    parent_cmd = sh.cmd("parent")
    child_kwargs = {
        "stdin": None,
        "input": b"test",
        "stdout": None,
        "stderr": None,
        "capture_output": False,
        "cwd": "/",
        "timeout": 10,
        "check": False,
        "encoding": "utf-8",
        "errors": "ignore",
        "text": False,
        "env": {"A": "B"},
        "replace_env": True,
    }
    child_popen_kwargs = {"umask": 1}
    child = parent_cmd.pipe("child", "cmd", **child_kwargs, **child_popen_kwargs)

    assert child.args == ["child", "cmd"]
    for attr, value in child_kwargs.items():
        assert getattr(child, attr) == value
    assert child.popen_kwargs == child_popen_kwargs
    assert child.parent.command is parent_cmd


@parametrize(
    "cmd, expected_shell_cmd",
    [
        param(sh.cmd("ls"), "ls"),
        param(sh.cmd("ps").pipe("grep", "foo bar"), "ps | grep 'foo bar'"),
        param(
            sh.cmd("ps").pipe("grep", "foo bar").pipe("grep", "test"),
            "ps | grep 'foo bar' | grep test",
        ),
        param(sh.cmd("cmd1").and_("cmd2", "a b").and_("cmd3"), "cmd1 && cmd2 'a b' && cmd3"),
        param(sh.cmd("cmd1").or_("cmd2", "a b").or_("cmd3"), "cmd1 || cmd2 'a b' || cmd3"),
        param(sh.cmd("cmd1").after("cmd2", "a b").after("cmd3"), "cmd1 ; cmd2 'a b' ; cmd3"),
        param(
            sh.cmd("cmd1").pipe("cmd2", "a b").and_("cmd3").or_("cmd4").after("cmd5"),
            "cmd1 | cmd2 'a b' && cmd3 || cmd4 ; cmd5",
        ),
    ],
)
def test_command_shell_cmd__returns_full_chained_command(cmd: sh.Command, expected_shell_cmd: str):
    assert cmd.shell_cmd == expected_shell_cmd



from pathlib import Path
import typing as t

import pytest
from pytest import param

import shelmet as sh

from .utils import Dir, File


parametrize = pytest.mark.parametrize


@parametrize(
    "files, pattern, expected_size",
    [
        param([File("a", size=10)], None, 10),
        param(
            [
                File("a", size=10),
                File("b/1", size=5),
                File("b/2", size=5),
                File("b/3", size=3),
                File("b/4", size=2),
                File("b/c/5", size=100),
                File("d", size=50),
            ],
            None,
            175,
        ),
        param(
            [
                File("a.json", size=123),
                File("b.txt", size=17),
                File("c.json", size=38),
                File("d", size=173),
            ],
            "*.json",
            161,
        ),
        param(
            [
                File("1/a.py", size=123),
                File("1/2/b.py", size=17),
                File("1/2/3/c.py", size=38),
                File("d.py", size=173),
                File("foo.txt", size=12),
                File("1/bar.txt", size=293),
                File("1/2/baz.txt", size=314),
                File("1/2/3/qux.txt", size=83),
            ],
            "**/*.py",
            351,
        ),
    ],
)
def test_getdirsize(
    tmp_path: Path, files: t.List[File], pattern: t.Optional[str], expected_size: int
):
    Dir(tmp_path, *files).mkdir()
    kwargs = {}
    if pattern:
        kwargs["pattern"] = pattern
    assert sh.getdirsize(tmp_path, **kwargs) == expected_size



import os

import pytest
from pytest import param

import shelmet as sh

from .utils import is_subdict


parametrize = pytest.mark.parametrize


@parametrize(
    "env",
    [
        param({"a": "1"}),
        param({"a": "1", "b": "2"}),
    ],
)
def test_environ__extends_envvars_and_restore_original(env: dict):
    orig_env = os.environ.copy()

    with sh.environ(env) as envvars:
        assert is_subdict(env, envvars)
        assert is_subdict(env, dict(os.environ))
        assert os.environ != orig_env
    assert os.environ == orig_env


@parametrize(
    "env",
    [
        param({"a": "1"}),
        param({"a": "1", "b": "2"}),
    ],
)
def test_environ__replaces_envvars_and_restores_original(env: dict):
    orig_env = os.environ.copy()

    with sh.environ(env, replace=True) as envvars:
        assert env == envvars
        assert env == os.environ
    assert os.environ == orig_env



import typing as t
from unittest import mock

import pytest

import shelmet as sh


parametrize = pytest.mark.parametrize


@pytest.fixture()
def mock_command() -> t.Generator[mock.MagicMock, None, None]:
    with mock.patch.object(sh.command, "Command", spec=sh.command.Command) as _mock_command:
        yield _mock_command


def test_run__creates_command_and_call_run(mock_command):
    args = ["ls", "-la"]
    kwargs: t.Dict[str, t.Any] = {
        "stdin": None,
        "input": "test",
        "stdout": None,
        "stderr": None,
        "capture_output": False,
        "combine_output": True,
        "cwd": "/",
        "timeout": 10,
        "check": False,
        "encoding": "utf-8",
        "errors": "ignore",
        "text": False,
        "env": {"A": "B"},
        "replace_env": True,
    }
    popen_kwargs = {"umask": 1}

    sh.run(*args, **kwargs, **popen_kwargs)

    assert mock_command.called
    assert mock_command.call_args == mock.call(*args, **kwargs, **popen_kwargs)

    mock_run: mock.MagicMock = mock_command.return_value.run
    assert mock_run.called
    assert mock_run.call_args == mock.call()



"""The fileio module contains utilities for file IO."""

from contextlib import contextmanager
import errno
from functools import partial
import io
import os
from pathlib import Path
import typing as t

from .filesystem import _candidate_temp_pathname, dirsync, fsync, mkdir, rm
from .types import (
    READ_ONLY_MODES,
    WRITE_ONLY_BIN_MODES,
    WRITE_ONLY_MODES,
    WRITE_ONLY_TEXT_MODES,
    ReadOnlyBinMode,
    ReadOnlyTextMode,
    StrPath,
    WriteOnlyBinMode,
    WriteOnlyTextMode,
)


DEFAULT_CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE


@contextmanager
def atomicdir(dir: StrPath, *, skip_sync: bool = False, overwrite: bool = True) -> t.Iterator[Path]:
    """
    Context-manager that is used to atomically create a directory and its contents.

    This context-manager will create a temporary directory in the same directory as the destination
    and yield the temporary directory as a ``pathblib.Path`` object. All atomic file system updates
    to the directory should then be done within the context-manager. Once the context-manager exits,
    the temporary directory will be passed to :func:`dirsync` (unless ``skip_sync=True``) and then
    moved to the destination followed by :func:`dirsync` on the parent directory. If the
    destination directory exists, it will be overwritten unless ``overwrite=False``.

    Args:
        dir: Directory path to create.
        skip_sync: Whether to skip calling :func:`dirsync` on the directory. Skipping this can help
            with performance at the cost of durability.
        overwrite: Whether to raise an exception if the destination exists once the directory is to
            be moved to its destination.
    """
    dst = Path(dir).absolute()
    if dst.is_file():
        raise FileExistsError(errno.EEXIST, f"Atomic directory target must not be a file: {dst}")

    tmp_dir = _candidate_temp_pathname(path=dst, prefix="_", suffix="_tmp")
    mkdir(tmp_dir)

    try:
        yield Path(tmp_dir)

        if not skip_sync:
            dirsync(tmp_dir)

        if overwrite:
            rm(dst)
        elif dst.exists():
            raise FileExistsError(
                errno.EEXIST,
                f"Atomic directory target must not exist when overwrite disabled: {dst}",
            )

        os.rename(tmp_dir, dst)

        if not skip_sync:
            dirsync(dst)
    finally:
        # In case something went wrong that prevented moving tmp_dir to dst.
        rm(tmp_dir)


@contextmanager
def atomicfile(
    file: StrPath,
    mode: str = "w",
    *,
    skip_sync: bool = False,
    overwrite: bool = True,
    **open_kwargs: t.Any,
) -> t.Iterator[t.IO]:
    """
    Context-manager similar to ``open()`` that is used to perform an atomic file write operation by
    first writing to a temporary location in the same directory as the destination and then renaming
    the file to the destination after all write operations are finished.

    This context-manager will open a temporary file for writing in the same directory as the
    destination and yield a file object just like ``open()`` does. All file operations while the
    context-manager is opened will be performed on the temporary file. Once the context-manager
    exits, the temporary file will flushed and fsync'd (unless ``skip_sync=True``). If the
    destination file exists, it will be overwritten unless ``overwrite=False``.

    Args:
        file: File path to write to.
        mode: File open mode.
        skip_sync: Whether to skip calling ``fsync`` on file. Skipping this can help with
            performance at the cost of durability.
        overwrite: Whether to raise an exception if the destination file exists once the file is to
            be written to its destination.
        **open_kwargs: Additional keyword arguments to ``open()`` when creating the temporary write
            file.
    """
    if isinstance(mode, str) and "x" in mode:
        raise ValueError(
            "Atomic file write mode 'x' is not supported. Use 'overwrite=False' instead."
        )

    if not isinstance(mode, str) or "w" not in mode:
        raise ValueError(f"Invalid atomic write mode: {mode}")

    dst = Path(file).absolute()
    if dst.is_dir():
        raise IsADirectoryError(errno.EISDIR, f"Atomic file target must not be a directory: {dst}")

    mkdir(dst.parent)
    tmp_file = _candidate_temp_pathname(path=dst, prefix="_", suffix=".tmp")

    try:
        with open(tmp_file, mode, **open_kwargs) as fp:
            yield fp
            if not skip_sync:
                fsync(fp)

        if overwrite:
            os.rename(tmp_file, dst)
        else:
            # This will fail if dst exists.
            os.link(tmp_file, dst)
            rm(tmp_file)

        if not skip_sync:
            dirsync(dst.parent)
    finally:
        # In case something went wrong that prevented moving tmp_file to dst.
        rm(tmp_file)


@t.overload
def read(file: StrPath, mode: ReadOnlyTextMode, **open_kwargs: t.Any) -> str:
    ...  # pragma: no cover


@t.overload
def read(file: StrPath, mode: ReadOnlyBinMode, **open_kwargs: t.Any) -> bytes:
    ...  # pragma: no cover


@t.overload
def read(file: StrPath, mode: str = "r", **open_kwargs: t.Any) -> t.Union[str, bytes]:
    ...  # pragma: no cover


def read(file: StrPath, mode: str = "r", **open_kwargs: t.Any) -> t.Union[str, bytes]:
    """
    Return contents of file.

    Args:
        file: File to read.
        mode: File open mode.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    if mode not in READ_ONLY_MODES:
        raise ValueError(f"Invalid read-only mode: {mode}")

    with open(file, mode, **open_kwargs) as fp:
        return fp.read()


def readbytes(file: StrPath, **open_kwargs: t.Any) -> bytes:
    """
    Return binary contents of file.

    Equivalent to calling :func:`read` with ``mode="rb"``.

    Args:
        file: File to read.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    return read(file, "rb", **open_kwargs)


def readtext(file: StrPath, **open_kwargs: t.Any) -> str:
    """
    Return text contents of file.

    Equivalent to calling :func:`read` with ``mode="r"`` (the default behavior of :func:`read`).

    Args:
        file: File to read.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    return read(file, "r", **open_kwargs)


@t.overload
def readchunks(
    file: StrPath,
    mode: ReadOnlyTextMode,
    *,
    size: int = ...,
    sep: t.Optional[str] = ...,
    **open_kwargs: t.Any,
) -> t.Generator[str, None, None]:
    ...  # pragma: no cover


@t.overload
def readchunks(
    file: StrPath,
    mode: ReadOnlyBinMode,
    *,
    size: int = ...,
    sep: t.Optional[bytes] = ...,
    **open_kwargs: t.Any,
) -> t.Generator[bytes, None, None]:
    ...  # pragma: no cover


@t.overload
def readchunks(
    file: StrPath,
    mode: str = "r",
    *,
    size: int = ...,
    sep: t.Optional[t.Union[str, bytes]] = ...,
    **open_kwargs: t.Any,
) -> t.Generator[t.Union[str, bytes], None, None]:
    ...  # pragma: no cover


def readchunks(
    file: StrPath,
    mode: str = "r",
    *,
    size: int = DEFAULT_CHUNK_SIZE,
    sep: t.Optional[t.Union[str, bytes]] = None,
    **open_kwargs: t.Any,
) -> t.Generator[t.Union[str, bytes], None, None]:
    """
    Yield contents of file as chunks.

    If separator, `sep`, is not given, chunks will be yielded by `size`.

    If separator, `sep`, is given, chunks will be yielded from as if from ``contents.split(sep)``.
    The `size` argument will still be used for each file read operation, but the contents will be
    buffered until a separator is encountered.

    Args:
        file: File to read.
        mode: File open mode.
        size: Size of chunks to read from file at a time and chunk size to yield when `sep` not
            given.
        sep: Separator to split chunks by in lieu of splitting by size.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    if mode not in READ_ONLY_MODES:
        raise ValueError(f"Invalid read-only mode: {mode}")
    return _readchunks(file, mode, size=size, sep=sep, **open_kwargs)


def _readchunks(file, mode="r", *, size=DEFAULT_CHUNK_SIZE, sep=None, **open_kwargs):
    buffer = ""
    if "b" in mode:
        buffer = b""

    with open(file, mode, **open_kwargs) as fp:
        try:
            while True:
                chunk = fp.read(size)

                if not chunk:
                    # We're done with the file but if we have anything in the buffer, yield it.
                    if buffer:
                        yield buffer
                    break
                elif not sep:
                    # Yield chunks delineated by size.
                    yield chunk
                else:
                    buffer += chunk
                    # Yield chunks delineated by separator.
                    while sep in buffer:
                        chunk, buffer = buffer.split(sep, 1)
                        yield chunk

        except GeneratorExit:  # pragma: no cover
            # Catch GeneratorExit to ensure contextmanager closes file when exiting generator early.
            pass


@t.overload
def readlines(
    file: StrPath, mode: ReadOnlyTextMode, *, limit: int = ..., **open_kwargs: t.Any
) -> t.Generator[str, None, None]:
    ...  # pragma: no cover


@t.overload
def readlines(
    file: StrPath, mode: ReadOnlyBinMode, *, limit: int = ..., **open_kwargs: t.Any
) -> t.Generator[bytes, None, None]:
    ...  # pragma: no cover


@t.overload
def readlines(
    file: StrPath, mode: str = "r", *, limit: int = ..., **open_kwargs: t.Any
) -> t.Generator[t.Union[str, bytes], None, None]:
    ...  # pragma: no cover


def readlines(
    file: StrPath, mode: str = "r", *, limit: int = -1, **open_kwargs: t.Any
) -> t.Generator[t.Union[str, bytes], None, None]:
    """
    Yield each line of a file.

    Note:
        Line-endings are included in the yielded values.

    Args:
        file: File to read.
        mode: File open mode.
        limit: Maximum length of each line to yield. For example, ``limit=10`` will yield the first
            10 characters of each line.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    if mode not in READ_ONLY_MODES:
        raise ValueError(f"Invalid read-only mode: {mode}")
    return _readlines(file, mode, limit=limit, **open_kwargs)


def _readlines(file, mode="r", *, limit=-1, **open_kwargs):
    sentinel = ""
    if "b" in mode:
        sentinel = b""

    with open(file, mode, **open_kwargs) as fp:
        try:
            yield from iter(lambda: fp.readline(limit), sentinel)
        except GeneratorExit:  # pragma: no cover
            # Catch GeneratorExit to ensure contextmanager closes file when exiting generator early.
            pass


@t.overload
def write(
    file: StrPath,
    contents: str,
    mode: WriteOnlyTextMode,
    *,
    atomic: bool = ...,
    **open_kwargs: t.Any,
) -> None:
    ...  # pragma: no cover


@t.overload
def write(
    file: StrPath,
    contents: bytes,
    mode: WriteOnlyBinMode,
    *,
    atomic: bool = ...,
    **open_kwargs: t.Any,
) -> None:
    ...  # pragma: no cover


@t.overload
def write(
    file: StrPath,
    contents: t.Union[str, bytes],
    mode: str = "w",
    *,
    atomic: bool = ...,
    **open_kwargs: t.Any,
) -> None:
    ...  # pragma: no cover


def write(
    file: StrPath,
    contents: t.Union[str, bytes],
    mode: str = "w",
    *,
    atomic: bool = False,
    **open_kwargs: t.Any,
) -> None:
    """
    Write contents to file.

    Args:
        file: File to write.
        contents: Contents to write.
        mode: File open mode.
        atomic: Whether to write the file to a temporary location in the same directory before
            moving it to the destination.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    if mode not in WRITE_ONLY_MODES:
        raise ValueError(f"Invalid write-only mode: {mode}")

    opener = open
    if atomic:
        overwrite = "x" not in mode
        mode = mode.replace("x", "w")
        opener = partial(atomicfile, overwrite=overwrite)  # type: ignore

    with opener(file, mode, **open_kwargs) as fp:
        fp.write(contents)


def writetext(
    file: StrPath, contents: str, mode: str = "w", *, atomic: bool = False, **open_kwargs: t.Any
) -> None:
    """
    Write text contents to file.

    Args:
        file: File to write.
        contents: Contents to write.
        mode: File open mode.
        atomic: Whether to write the file to a temporary location in the same directory before
            moving it to the destination.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    if mode not in WRITE_ONLY_TEXT_MODES:
        raise ValueError(f"Invalid write-only text-mode: {mode}")
    write(file, contents, mode, atomic=atomic, **open_kwargs)


def writebytes(
    file: StrPath, contents: bytes, mode: str = "wb", *, atomic: bool = False, **open_kwargs: t.Any
) -> None:
    """
    Write binary contents to file.

    Args:
        file: File to write.
        contents: Contents to write.
        mode: File open mode.
        atomic: Whether to write the file to a temporary location in the same directory before
            moving it to the destination.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    if mode not in WRITE_ONLY_BIN_MODES:
        raise ValueError(f"Invalid write-only binary-mode: {mode}")
    write(file, contents, mode, atomic=atomic, **open_kwargs)


@t.overload
def writelines(
    file: StrPath,
    items: t.Iterable[str],
    mode: WriteOnlyTextMode,
    *,
    ending: t.Optional[str] = None,
    atomic: bool = False,
    **open_kwargs: t.Any,
) -> None:
    ...  # pragma: no cover


@t.overload
def writelines(
    file: StrPath,
    items: t.Iterable[bytes],
    mode: WriteOnlyBinMode,
    *,
    ending: t.Optional[bytes] = None,
    atomic: bool = False,
    **open_kwargs: t.Any,
) -> None:
    ...  # pragma: no cover


@t.overload
def writelines(
    file: StrPath,
    items: t.Union[t.Iterable[str], t.Iterable[bytes]],
    mode: str = "w",
    *,
    ending: t.Optional[t.Union[str, bytes]] = None,
    atomic: bool = False,
    **open_kwargs: t.Any,
) -> None:
    ...  # pragma: no cover


def writelines(
    file: StrPath,
    items: t.Union[t.Iterable[str], t.Iterable[bytes]],
    mode: str = "w",
    *,
    ending: t.Optional[t.Union[str, bytes]] = None,
    atomic: bool = False,
    **open_kwargs: t.Any,
) -> None:
    """
    Write lines to file.

    Args:
        file: File to write.
        items: Items to write.
        mode: File open mode.
        ending: Line ending to use. Defaults to newline.
        atomic: Whether to write the file to a temporary location in the same directory before
            moving it to the destination.
        **open_kwargs: Additional keyword arguments to pass to ``open``.
    """
    if mode not in WRITE_ONLY_MODES:
        raise ValueError(f"Invalid write-only mode: {mode}")

    if ending is None:
        ending = "\n"
        if "b" in mode:
            ending = b"\n"

    opener = open
    if atomic:
        overwrite = "x" not in mode
        mode = mode.replace("x", "w")
        opener = partial(atomicfile, overwrite=overwrite)  # type: ignore

    lines = (item + ending for item in items)  # type: ignore
    with opener(file, mode, **open_kwargs) as fp:
        fp.writelines(lines)



from pathlib import Path
import typing as t
from unittest import mock

import pytest
from pytest import param

import shelmet as sh


parametrize = pytest.mark.parametrize


@pytest.fixture(
    params=[
        param("w"),
        param("wt"),
        param("tw"),
        param("a"),
        param("at"),
        param("ta"),
        param("x"),
        param("xt"),
        param("tx"),
        param("wb"),
        param("bw"),
        param("ab"),
        param("ba"),
        param("xb"),
        param("bx"),
    ]
)
def valid_write_only_mode(request) -> str:
    return request.param


@pytest.fixture(
    params=[
        param("wb"),
        param("bw"),
        param("ab"),
        param("ba"),
        param("xb"),
        param("bx"),
    ]
)
def valid_write_only_bin_mode(request) -> str:
    return request.param


@pytest.fixture(
    params=[
        param("w"),
        param("wt"),
        param("tw"),
        param("a"),
        param("at"),
        param("ta"),
        param("x"),
        param("xt"),
        param("tx"),
    ]
)
def valid_write_only_text_mode(request) -> str:
    return request.param


@pytest.fixture(
    params=[
        param("r"),
        param("r+"),
        param("rb"),
        param("rb+"),
        param("w+"),
        param("wb+"),
        param("a+"),
        param("ab+"),
    ]
)
def invalid_write_only_mode(request) -> str:
    return request.param


@pytest.fixture(
    params=[
        param("r"),
        param("r+"),
        param("rb"),
        param("rb+"),
        param("w+"),
        param("wb+"),
        param("a+"),
        param("ab+"),
        param("w"),
        param("wt"),
        param("a"),
        param("x"),
    ]
)
def invalid_write_only_bin_mode(request) -> str:
    return request.param


@pytest.fixture(
    params=[
        param("r"),
        param("r+"),
        param("rb"),
        param("rb+"),
        param("w+"),
        param("wb+"),
        param("a+"),
        param("ab+"),
        param("wb"),
        param("ab"),
        param("xb"),
    ]
)
def invalid_write_only_text_mode(request) -> str:
    return request.param


@pytest.fixture()
def mock_atomicfile():
    with mock.patch("shelmet.fileio.atomicfile") as _mock_atomicfile:
        yield _mock_atomicfile


@parametrize(
    "mode, contents",
    [
        param("w", "abcd"),
        param("a", "abcd"),
        param("x", "abcd"),
        param("wb", b"abcd"),
        param("ab", b"abcd"),
        param("xb", b"abcd"),
    ],
)
def test_write(tmp_path: Path, mode: str, contents: t.Union[str, bytes]):
    file = tmp_path / "test_file"
    sh.write(file, contents, mode)

    actual_contents = file.read_bytes()
    if isinstance(contents, str):
        actual_contents = actual_contents.decode()  # type: ignore

    assert contents == actual_contents


def test_write__accepts_valid_mode(tmp_path: Path, valid_write_only_mode: str):
    contents: t.Union[str, bytes] = b"" if "b" in valid_write_only_mode else ""
    sh.write(tmp_path / "test_file", contents, valid_write_only_mode)


@parametrize(
    "mode, contents, expected_mode, expected_overwrite",
    [
        param("w", "test", "w", True),
        param("wb", b"test", "wb", True),
        param("x", "test", "w", False),
        param("xb", b"test", "wb", False),
    ],
)
def test_write__can_atomically_create_file(
    tmp_path: Path,
    mock_atomicfile: mock.MagicMock,
    mode: str,
    contents: t.Union[str, bytes],
    expected_mode: str,
    expected_overwrite: bool,
):
    file = tmp_path / "test_file"
    sh.write(file, contents, mode, atomic=True)

    assert mock_atomicfile.called
    assert mock_atomicfile.call_args == mock.call(file, expected_mode, overwrite=expected_overwrite)

    args, kwargs = mock_atomicfile.call_args
    with mock_atomicfile(*args, **kwargs) as fp:
        assert fp.write.called
        assert fp.write.call_args == mock.call(contents)


def test_write__raises_when_mode_invalid(tmp_path: Path, invalid_write_only_mode: str):
    file = tmp_path / "test_file"
    with pytest.raises(ValueError):
        sh.write(file, "", invalid_write_only_mode)


@parametrize(
    "mode, contents",
    [
        param("wb", b"abcd"),
        param("ab", b"abcd"),
        param("xb", b"abcd"),
    ],
)
def test_writebytes(tmp_path: Path, mode: str, contents: bytes):
    file = tmp_path / "test_file"
    sh.writebytes(file, contents, mode)

    actual_contents = file.read_bytes()
    assert contents == actual_contents


def test_writebytes__accepts_valid_mode(tmp_path: Path, valid_write_only_bin_mode: str):
    sh.write(tmp_path / "test_file", b"", valid_write_only_bin_mode)


def test_writebytes__raises_when_mode_invalid(tmp_path: Path, invalid_write_only_bin_mode: str):
    file = tmp_path / "test_file"
    with pytest.raises(ValueError):
        sh.writebytes(file, b"", invalid_write_only_bin_mode)


@parametrize(
    "mode, contents",
    [
        param("w", "abcd"),
        param("a", "abcd"),
        param("x", "abcd"),
    ],
)
def test_writetext(tmp_path: Path, mode: str, contents: str):
    file = tmp_path / "test_file"
    sh.writetext(file, contents, mode)

    actual_contents = file.read_text()
    assert contents == actual_contents


def test_writetext__accepts_valid_mode(tmp_path: Path, valid_write_only_text_mode: str):
    sh.write(tmp_path / "test_file", "", valid_write_only_text_mode)


def test_writetext__raises_when_mode_invalid(tmp_path: Path, invalid_write_only_text_mode: str):
    file = tmp_path / "test_file"
    with pytest.raises(ValueError):
        sh.writetext(file, "", invalid_write_only_text_mode)


@parametrize(
    "mode, items",
    [
        param("w", ["a", "b", "c"]),
        param("a", ["a", "b", "c"]),
        param("x", ["a", "b", "c"]),
        param("w", ["a", "b", "c"]),
        param("a", ["a", "b", "c"]),
        param("x", ["a", "b", "c"]),
        param("wb", [b"a", b"b", b"c"]),
        param("ab", [b"a", b"b", b"c"]),
        param("xb", [b"a", b"b", b"c"]),
        param("wb", [b"a", b"b", b"c"]),
        param("ab", [b"a", b"b", b"c"]),
        param("xb", [b"a", b"b", b"c"]),
    ],
)
def test_writelines(tmp_path: Path, mode: str, items: t.List[t.AnyStr]):
    file = tmp_path / "test_file"
    sh.writelines(file, items, mode)

    read_mode = "r"
    if "b" in mode:
        read_mode += "b"

    with open(file, read_mode) as fp:
        lines = fp.readlines()

    for i, line in enumerate(lines):
        assert items[i] == line.strip()


def test_writelines__accepts_valid_mode(tmp_path: Path, valid_write_only_mode: str):
    contents: t.Union[str, bytes] = b"" if "b" in valid_write_only_mode else ""
    sh.writelines(tmp_path / "test_file", [contents], valid_write_only_mode)  # type: ignore


@parametrize(
    "mode, items, ending",
    [
        param("w", ["a", "b", "c"], "|"),
        param("a", ["a", "b", "c"], "|"),
        param("x", ["a", "b", "c"], "|"),
        param("w", ["a", "b", "c"], "|"),
        param("a", ["a", "b", "c"], "|"),
        param("x", ["a", "b", "c"], "|"),
        param("wb", [b"a", b"b", b"c"], b"|"),
        param("ab", [b"a", b"b", b"c"], b"|"),
        param("xb", [b"a", b"b", b"c"], b"|"),
        param("wb", [b"a", b"b", b"c"], b"|"),
        param("ab", [b"a", b"b", b"c"], b"|"),
        param("xb", [b"a", b"b", b"c"], b"|"),
    ],
)
def test_writelines__uses_custom_ending(
    tmp_path: Path, mode: str, items: t.List[t.AnyStr], ending: t.AnyStr
):
    file = tmp_path / "test_file"
    sh.writelines(file, items, mode, ending=ending)

    read_mode = "r"
    if "b" in mode:
        read_mode += "b"

    with open(file, read_mode) as fp:
        contents = fp.read()

    actual_items = contents.rstrip(ending).split(ending)
    for i, actual_item in enumerate(actual_items):
        assert items[i] == actual_item


@parametrize(
    "mode, items, expected_mode, expected_overwrite",
    [
        param("w", ["test"], "w", True),
        param("wb", [b"test"], "wb", True),
        param("x", ["test"], "w", False),
        param("xb", [b"test"], "wb", False),
    ],
)
def test_writelines__can_atomically_create_file(
    tmp_path: Path,
    mock_atomicfile: mock.MagicMock,
    mode: str,
    items: t.Union[t.List[str], t.List[bytes]],
    expected_mode: str,
    expected_overwrite: bool,
):
    file = tmp_path / "test_file"
    sh.writelines(file, items, mode, atomic=True)

    assert mock_atomicfile.called
    assert mock_atomicfile.call_args == mock.call(file, expected_mode, overwrite=expected_overwrite)

    args, kwargs = mock_atomicfile.call_args
    with mock_atomicfile(*args, **kwargs) as fp:
        assert fp.writelines.called
        lines = [line.strip() for line in fp.writelines.call_args[0][0]]
        assert items == lines


def test_writelines__raises_when_mode_invalid(tmp_path: Path, invalid_write_only_mode: str):
    file = tmp_path / "test_file"
    with pytest.raises(ValueError):
        sh.writelines(file, [], invalid_write_only_mode)



"""The command module contains utilities for working with subprocess commands."""

from abc import ABCMeta, abstractmethod
import os
import shlex
import subprocess
import typing as t
from typing import Iterable

from .types import RunArgs, StdFile, StrPath


class Command:
    """
    A system command that can be executed multiple times and used to create piped commands.

    Executing the command is done using :meth:`run` which is a wrapper around ``subprocess.run``.
    However, the default arguments for a :class:`Command` enable different default behavior than
    ``subprocess.run``:

    - Output is captured
    - Text-mode is enabled
    - Environment variables extend ``os.environ`` instead of replacing them.
    - Exceptions are raised by default when the completed process returns a non-zero exit code.
    - System command arguments can be passed as a var-args instead of just a list.

    To disable output capture completely, use ``capture_output=False``. To disable output capture
    for just one of them, set ``stdout`` or ``stderr`` to ``None``.

    To disable ``os.environ`` extension, use ``replace_env=True``.

    To disable exception raising, use ``check=False``.

    Therefore, to use the default behavior of ``subprocess.run``, set the following keyword
    arguments:

    ::

        ls = Command(["ls", "-la"], capture_output=False, text=False, check=False, replace_env=True)
        ls.run()

    Args:
        *args: System command arguments to execute. If ``None`` is given as an argument value, it
            will be discarded.
        stdin: Specify the executed command’s standard input.
        input: If given it will be passed to the underlying process as stdin. When used, stdin will
            be set to ``PIPE`` automatically and cannot be used. The value will be encoded or
            decoded automatically if it does not match the expected type based on whether text-mode
            is enabled or not.
        stdout: Specify the executed command’s standard output.
        stderr: Specify the executed command’s standard error.
        capture_output: Whether to capture stdout and stderr and include in the returned completed
            process result.
        combine_output: Whether to combine stdout and stderr. Equilvalent to setting
            ``stderr=subprocess.STDOUT``.
        cwd: Set the current working directory when executing the command.
        timeout: If the timeout expires, the child process will be killed and waited for.
        check: Whether to check return code and raise if it is non-zero.
        encoding: Set encoding to use for text-mode.
        errors: Specify how encoding and decoding errors should be handled. Must be one of "strict",
            "ignore", "replace", "backslashreplace", "xmlcharrefreplace", "namereplace"
        text: Set text-mode.
        env: Environment variables for the new process. Unlike in ``subprocess.run``, the default
            behavior is to extend the existing environment. Use ``replace_env=True`` to replace the
            environment variables instead.
        replace_env: Whether to replace the current environment when `env` given.

    Keyword Args:
        All other keyword arguments are passed to ``subprocess.run`` which subsequently passes them
        to ``subprocess.Popen``.
    """

    def __init__(
        self,
        *args: RunArgs,
        stdin: t.Optional[StdFile] = None,
        input: t.Optional[t.Union[str, bytes]] = None,
        stdout: t.Optional[StdFile] = subprocess.PIPE,
        stderr: t.Optional[StdFile] = subprocess.PIPE,
        capture_output: bool = True,
        combine_output: bool = False,
        cwd: t.Optional[StrPath] = None,
        timeout: t.Optional[t.Union[float, int]] = None,
        check: bool = True,
        encoding: t.Optional[str] = None,
        errors: t.Optional[str] = None,
        text: bool = True,
        env: t.Optional[dict] = None,
        replace_env: bool = False,
        parent: t.Optional["ChainCommand"] = None,
        **popen_kwargs: t.Any,
    ):
        run_args = _parse_run_args(args, error_prefix=f"{self.__class__.__name__}(): ")

        if input is not None:
            # Coerce input based on text mode setting.
            if text and isinstance(input, bytes):
                input = input.decode()
            elif not text and isinstance(input, str):
                input = input.encode()

        if not capture_output:
            stdout = None
            stderr = None

        if combine_output:
            stderr = subprocess.STDOUT

        self.args = run_args
        self.stdin = stdin
        self.input = input
        self.stdout = stdout
        self.stderr = stderr
        self.capture_output = capture_output
        self.combine_output = combine_output
        self.cwd = cwd
        self.timeout = timeout
        self.check = check
        self.encoding = encoding
        self.errors = errors
        self.text = text
        self.env = env
        self.replace_env = replace_env
        self.popen_kwargs = popen_kwargs
        self.parent = parent

    @property
    def parents(self) -> t.List["ChainCommand"]:
        """Return list of parent :class:`Command` objects that pipe output into this command."""
        parents = []
        if self.parent:
            grand_parents = self.parent.command.parents
            if grand_parents:
                parents.extend(grand_parents)
            parents.append(self.parent)
        return parents

    @property
    def shell_cmd(self) -> str:
        """Return string version of command that would be used when executing from a shell."""
        cmd = " ".join(
            shlex.quote(arg.decode() if isinstance(arg, bytes) else arg) for arg in self.args
        )
        if self.parent:
            cmd = f"{self.parent.shell_cmd} {cmd}"
        return cmd

    def __repr__(self) -> str:
        kv_items: t.List[t.Tuple[str, t.Any]] = [("args", self.args)]

        parents = self.parents
        if parents:
            repr_parents = ", ".join(repr(parent) for parent in self.parents)
            kv_items.append(("parents", f"[{repr_parents}]"))

        kv_out = ", ".join(f"{key}={value}" for key, value in kv_items)
        return f"{self.__class__.__name__}({kv_out})"

    @classmethod
    def from_command(cls, command, *extra_args: RunArgs, **override_kwargs: t.Any) -> "Command":
        run_args = command.args
        if extra_args:
            run_args = run_args + _parse_run_args(
                extra_args, error_prefix=f"{cls.__name__}.run(): "
            )

        override_kwargs.setdefault("input", command.input)
        override_kwargs.setdefault("stdin", command.stdin)
        override_kwargs.setdefault("stdout", command.stdout)
        override_kwargs.setdefault("stderr", command.stderr)
        override_kwargs.setdefault("capture_output", command.capture_output)
        override_kwargs.setdefault("combine_output", command.combine_output)
        override_kwargs.setdefault("cwd", command.cwd)
        override_kwargs.setdefault("timeout", command.timeout)
        override_kwargs.setdefault("check", command.check)
        override_kwargs.setdefault("encoding", command.encoding)
        override_kwargs.setdefault("errors", command.errors)
        override_kwargs.setdefault("text", command.text)
        override_kwargs.setdefault("env", command.env)
        override_kwargs.setdefault("replace_env", command.replace_env)
        override_kwargs.update(command.popen_kwargs)

        return cls(*run_args, **override_kwargs)

    def pipe(
        self,
        *args: RunArgs,
        stdin: t.Optional[StdFile] = None,
        input: t.Optional[t.Union[str, bytes]] = None,
        stdout: t.Optional[StdFile] = subprocess.PIPE,
        stderr: t.Optional[StdFile] = subprocess.PIPE,
        capture_output: bool = True,
        combine_output: bool = False,
        cwd: t.Optional[StrPath] = None,
        timeout: t.Optional[t.Union[float, int]] = None,
        check: bool = True,
        encoding: t.Optional[str] = None,
        errors: t.Optional[str] = None,
        text: bool = True,
        env: t.Optional[dict] = None,
        replace_env: bool = False,
        **popen_kwargs: t.Any,
    ) -> "Command":
        """
        Return a new command whose input will be piped from the output of this command.

        This is like running "<this-command> | <next-command>".
        """
        return self.__class__(
            *args,
            stdin=stdin,
            input=input,
            stdout=stdout,
            stderr=stderr,
            capture_output=capture_output,
            combine_output=combine_output,
            cwd=cwd,
            timeout=timeout,
            check=check,
            encoding=encoding,
            errors=errors,
            text=text,
            env=env,
            replace_env=replace_env,
            parent=PipeCommand(self),
            **popen_kwargs,
        )

    def after(
        self,
        *args: RunArgs,
        stdin: t.Optional[StdFile] = None,
        input: t.Optional[t.Union[str, bytes]] = None,
        stdout: t.Optional[StdFile] = subprocess.PIPE,
        stderr: t.Optional[StdFile] = subprocess.PIPE,
        capture_output: bool = True,
        combine_output: bool = False,
        cwd: t.Optional[StrPath] = None,
        timeout: t.Optional[t.Union[float, int]] = None,
        check: bool = True,
        encoding: t.Optional[str] = None,
        errors: t.Optional[str] = None,
        text: bool = True,
        env: t.Optional[dict] = None,
        replace_env: bool = False,
        **popen_kwargs: t.Any,
    ) -> "Command":
        """
        Return a new command that will be executed after this command regardless of this command's
        return code.

        This is like running "<this-command> ; <next-command>".
        """
        return self.__class__(
            *args,
            stdin=stdin,
            input=input,
            stdout=stdout,
            stderr=stderr,
            capture_output=capture_output,
            combine_output=combine_output,
            cwd=cwd,
            timeout=timeout,
            check=check,
            encoding=encoding,
            errors=errors,
            text=text,
            env=env,
            replace_env=replace_env,
            parent=AfterCommand(self),
            **popen_kwargs,
        )

    def and_(
        self,
        *args: RunArgs,
        stdin: t.Optional[StdFile] = None,
        input: t.Optional[t.Union[str, bytes]] = None,
        stdout: t.Optional[StdFile] = subprocess.PIPE,
        stderr: t.Optional[StdFile] = subprocess.PIPE,
        capture_output: bool = True,
        combine_output: bool = False,
        cwd: t.Optional[StrPath] = None,
        timeout: t.Optional[t.Union[float, int]] = None,
        check: bool = True,
        encoding: t.Optional[str] = None,
        errors: t.Optional[str] = None,
        text: bool = True,
        env: t.Optional[dict] = None,
        replace_env: bool = False,
        **popen_kwargs: t.Any,
    ) -> "Command":
        """
        Return a new command that will be AND'd with this command.

        This is like running "<this-command> && <next-command>".
        """
        return self.__class__(
            *args,
            stdin=stdin,
            input=input,
            stdout=stdout,
            stderr=stderr,
            capture_output=capture_output,
            combine_output=combine_output,
            cwd=cwd,
            timeout=timeout,
            check=check,
            encoding=encoding,
            errors=errors,
            text=text,
            env=env,
            replace_env=replace_env,
            parent=AndCommand(self),
            **popen_kwargs,
        )

    def or_(
        self,
        *args: RunArgs,
        stdin: t.Optional[StdFile] = None,
        input: t.Optional[t.Union[str, bytes]] = None,
        stdout: t.Optional[StdFile] = subprocess.PIPE,
        stderr: t.Optional[StdFile] = subprocess.PIPE,
        capture_output: bool = True,
        combine_output: bool = False,
        cwd: t.Optional[StrPath] = None,
        timeout: t.Optional[t.Union[float, int]] = None,
        check: bool = True,
        encoding: t.Optional[str] = None,
        errors: t.Optional[str] = None,
        text: bool = True,
        env: t.Optional[dict] = None,
        replace_env: bool = False,
        **popen_kwargs: t.Any,
    ) -> "Command":
        """
        Return a new command that will be OR'd with this command.

        This is like running "<this-command> || <next-command>".
        """
        return self.__class__(
            *args,
            stdin=stdin,
            input=input,
            stdout=stdout,
            stderr=stderr,
            capture_output=capture_output,
            combine_output=combine_output,
            cwd=cwd,
            timeout=timeout,
            check=check,
            encoding=encoding,
            errors=errors,
            text=text,
            env=env,
            replace_env=replace_env,
            parent=OrCommand(self),
            **popen_kwargs,
        )

    def run(self, *extra_args: RunArgs, **override_kwargs: t.Any) -> subprocess.CompletedProcess:
        """
        Wrapper around ``subprocess.run`` that uses this class' arguments as defaults.

        To add additional command args to :attr:`args`, pass them as var-args.

        To override default keyword arguments, pass them as keyword-args.

        If :attr:`parent` is set (e.g. if this command was created with :meth:`pipe`,
        :meth:`after`, :meth:`and_`, or :meth:`or_`), then the parent command will be called first
        and then chained with this command.

        Args:
            *extra_args: Extend :attr:`args` with extra command arguments.
            **override_kwargs: Override this command's keyword arguments.
        """
        if extra_args or override_kwargs or self.parent:
            command = self.from_command(self, *extra_args, **override_kwargs)
        else:
            command = self

        if self.parent:
            result = self.parent.run(command)
        else:
            result = command._run()

        return result

    def _run(self):
        if self.env and not self.replace_env:
            env = {**os.environ, **self.env}
        else:
            env = self.env

        return subprocess.run(
            self.args,
            stdin=self.stdin,
            input=self.input,
            stdout=self.stdout,
            stderr=self.stderr,
            cwd=self.cwd,
            timeout=self.timeout,
            check=self.check,
            encoding=self.encoding,
            errors=self.errors,
            universal_newlines=self.text,  # NOTE: "text" argument doesn't exist in Python 3.6.
            env=env,
            **self.popen_kwargs,
        )


class ChainCommand(metaclass=ABCMeta):
    """
    Abstract base class for representing a chained command.

    The class is initialized with a parent command that will be called before the next command in
    the chain. The next command will be passed into the :meth:`run` method. Each subclass is
    responsible for implementing the :meth:`run` logic.
    """

    def __init__(self, command: Command):
        self.command = command

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(args={self.command.args})"

    @property
    def shell_cmd(self):
        return f"{self.command.shell_cmd} {self.separator}"

    @property
    @abstractmethod
    def separator(self) -> str:  # pragma: no cover
        # The separator is used for representational purposes only in `shell_cmd`.
        pass

    @abstractmethod
    def run(self, next_command: Command) -> subprocess.CompletedProcess:  # pragma: no cover
        # The primary logic that runs the parent command followed by the next command if applicable.
        pass


class AfterCommand(ChainCommand):
    """
    Chained command that runs one command after another regardless of the first command's return
    code.

    This is like the shell equivalent of "<cmd1> ; <cmd2>".
    """

    separator = ";"

    def run(self, next_command: Command):
        """Run `next_command` after :attr:`command`."""
        try:
            self.command.run()
        except subprocess.CalledProcessError:
            pass
        return next_command.run()


class PipeCommand(ChainCommand):
    """
    Chained command that pipes the output of one command into the input of another.

    This is like the shell equivalent of "<cmd1> | <cmd2>".
    """

    separator = "|"

    def run(self, next_command: Command) -> subprocess.CompletedProcess:
        """Pipe :attr:`command` into `next_command`."""
        result: t.Union[subprocess.CompletedProcess, subprocess.CalledProcessError]
        try:
            result = self.command.run()
        except subprocess.CalledProcessError as exc:
            result = exc
        return next_command.run(input=result.stdout, stdin=None)


class AndCommand(ChainCommand):
    """
    Chained command that runs one command after the other if the first command succeeds.

    This is like the shell equivalent of "<cmd1> | <cmd2>".
    """

    separator = "&&"

    def run(self, next_command: Command) -> subprocess.CompletedProcess:
        """Run `next_command` after :attr:`command` if :attr:`command` succeeds."""
        result = self.command.run()
        if result.returncode == 0:
            result = next_command.run()
        return result


class OrCommand(ChainCommand):
    """
    Chained command that runs one command after the other if the first command fails.

    This is like the shell equivalent of "<cmd1> | <cmd2>".
    """

    separator = "||"

    def run(self, next_command: Command) -> subprocess.CompletedProcess:
        """Run `next_command` after :attr:`command` if :attr:`command` fails."""
        try:
            result = self.command.run()
        except subprocess.CalledProcessError:
            failed = True
        else:
            failed = result.returncode != 0
        if failed:
            result = next_command.run()
        return result


def _parse_run_args(args: tuple, error_prefix: str = "run(): ") -> t.List[t.Union[str, bytes]]:
    good_args = []
    bad_args = []

    for arg in _flatten(args):
        if arg is None:
            # Ignore None values.
            continue
        elif isinstance(arg, (str, bytes)):
            good_args.append(arg)
        else:
            bad_args.append(arg)

    if bad_args:
        raise TypeError(
            f"{error_prefix}requires all positional arguments to be either string or bytes, not"
            f" {bad_args}"
        )

    if not good_args:
        raise TypeError(f"{error_prefix}requires at least one non-empty positional argument")

    return good_args


def _flatten(items: t.Iterable) -> t.Generator[t.Any, None, None]:
    for item in items:
        if isinstance(item, Iterable) and not isinstance(item, (str, bytes)):
            yield from item
        else:
            yield item


def cmd(
    *args: RunArgs,
    stdin: t.Optional[StdFile] = None,
    input: t.Optional[t.Union[str, bytes]] = None,
    stdout: t.Optional[StdFile] = subprocess.PIPE,
    stderr: t.Optional[StdFile] = subprocess.PIPE,
    capture_output: bool = True,
    combine_output: bool = False,
    cwd: t.Optional[StrPath] = None,
    timeout: t.Optional[t.Union[float, int]] = None,
    check: bool = True,
    encoding: t.Optional[str] = None,
    errors: t.Optional[str] = None,
    text: bool = True,
    env: t.Optional[dict] = None,
    replace_env: bool = False,
    **popen_kwargs: t.Any,
) -> Command:
    """
    Factory that returns an instance of :class:`.Command` initialized with the given arguments.

    See Also:
        :class:`.Command` for description of arguments.
    """
    return Command(
        *args,
        stdin=stdin,
        input=input,
        stdout=stdout,
        stderr=stderr,
        capture_output=capture_output,
        combine_output=combine_output,
        cwd=cwd,
        timeout=timeout,
        check=check,
        encoding=encoding,
        errors=errors,
        text=text,
        env=env,
        replace_env=replace_env,
        **popen_kwargs,
    )


def run(
    *args: RunArgs,
    stdin: t.Optional[StdFile] = None,
    input: t.Optional[t.Union[str, bytes]] = None,
    stdout: t.Optional[StdFile] = subprocess.PIPE,
    stderr: t.Optional[StdFile] = subprocess.PIPE,
    capture_output: bool = True,
    combine_output: bool = False,
    cwd: t.Optional[StrPath] = None,
    timeout: t.Optional[t.Union[float, int]] = None,
    check: bool = True,
    encoding: t.Optional[str] = None,
    errors: t.Optional[str] = None,
    text: bool = True,
    env: t.Optional[dict] = None,
    replace_env: bool = False,
    **popen_kwargs: t.Any,
) -> subprocess.CompletedProcess:
    """
    Convenience function-wrapper around :meth:`.Command.run`.

    Using this function is equivalent to:

    ::

        result = sh.cmd(*args, **kwargs).run()

    See Also:
        :class:`.Command` for description of arguments.
    """
    cmd = Command(
        *args,
        stdin=stdin,
        input=input,
        stdout=stdout,
        stderr=stderr,
        capture_output=capture_output,
        combine_output=combine_output,
        cwd=cwd,
        timeout=timeout,
        check=check,
        encoding=encoding,
        errors=errors,
        text=text,
        env=env,
        replace_env=replace_env,
        **popen_kwargs,
    )
    return cmd.run()



"""The types module contains common type annotation definitions."""

from pathlib import Path
import typing as t

from typing_extensions import Literal


def _get_literal_args(literal_type) -> tuple:  # pragma: no cover
    """Backwards compatible method to get arguments passed to ``Literal`` in lieu of
    ``typing.get_args``."""
    if hasattr(t, "get_args"):
        # Python 3.8+
        # pylint: disable=no-member
        return t.get_args(literal_type)  # type: ignore
    elif hasattr(literal_type, "__args__"):
        # Python 3.7
        return literal_type.__args__
    else:
        # Python 3.6
        return literal_type.__values__


StrPath = t.Union[str, Path]
LsFilterFn = t.Callable[[Path], bool]
LsFilterable = t.Union[str, t.Pattern, LsFilterFn]
LsFilter = t.Union[LsFilterable, t.Iterable[LsFilterable]]
StdFile = t.Union[int, t.IO[t.Any]]
RunArgs = t.Union[str, bytes, None, t.Iterable[t.Union[str, bytes, None]]]
ReadOnlyTextMode = Literal["r", "rt", "tr"]
ReadOnlyBinMode = Literal["rb", "br"]
WriteOnlyTextMode = Literal["w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx"]
WriteOnlyBinMode = Literal["wb", "bw", "ab", "ba", "xb", "bx"]

READ_ONLY_TEXT_MODES = _get_literal_args(ReadOnlyTextMode)
READ_ONLY_BIN_MODES = _get_literal_args(ReadOnlyBinMode)
READ_ONLY_MODES = READ_ONLY_TEXT_MODES + READ_ONLY_BIN_MODES
WRITE_ONLY_TEXT_MODES = _get_literal_args(WriteOnlyTextMode)
WRITE_ONLY_BIN_MODES = _get_literal_args(WriteOnlyBinMode)
WRITE_ONLY_MODES = WRITE_ONLY_TEXT_MODES + WRITE_ONLY_BIN_MODES



import grp
import os
from pathlib import Path
import pwd
import typing as t
from unittest import mock
from uuid import uuid4

import pytest

import shelmet as sh

from .utils import Dir, File


parametrize = pytest.mark.parametrize


@pytest.fixture()
def os_user() -> pwd.struct_passwd:
    return pwd.getpwuid(os.getuid())


@pytest.fixture()
def os_group() -> grp.struct_group:
    return grp.getgrgid(os.getgid())


@pytest.fixture()
def test_file(tmp_path: Path) -> Path:
    test_file = tmp_path / "test_file.txt"
    test_file.touch()
    return test_file


@pytest.fixture()
def test_dir(tmp_path: Path) -> Path:
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    return test_dir


@pytest.fixture()
def mock_os_chown() -> t.Generator[mock.MagicMock, None, None]:
    with mock.patch("os.chown") as mocked_os_chown:
        yield mocked_os_chown


def chown_call(
    path: t.Union[str, Path, int], user: int = -1, group: int = -1, follow_symlinks: bool = True
) -> tuple:
    if isinstance(path, str):
        path = Path(path)
    return mock.call(path, user, group, follow_symlinks=follow_symlinks)


def test_chown__changes_ownership_given_uid(tmp_path: Path, mock_os_chown: mock.MagicMock):
    sh.chown(tmp_path, user=1)
    assert mock_os_chown.call_args == chown_call(tmp_path, user=1)


def test_chown__changes_ownership_given_gid(tmp_path: Path, mock_os_chown: mock.MagicMock):
    sh.chown(tmp_path, group=1)
    assert mock_os_chown.call_args == chown_call(tmp_path, group=1)


def test_chown__changes_ownership_given_user_name(
    tmp_path: Path, mock_os_chown: mock.MagicMock, os_user: pwd.struct_passwd
):
    sh.chown(tmp_path, user=os_user.pw_name)
    assert mock_os_chown.call_args == chown_call(tmp_path, user=os_user.pw_uid)


def test_chown__changes_ownership_given_group_name(
    tmp_path: Path, mock_os_chown: mock.MagicMock, os_group: grp.struct_group
):
    sh.chown(tmp_path, group=os_group.gr_name)
    assert mock_os_chown.call_args == chown_call(tmp_path, group=os_group.gr_gid)


def test_chown__changes_ownership_given_file_descriptor(mock_os_chown: mock.MagicMock):
    sh.chown(1, user=2, group=3)
    assert mock_os_chown.call_args == chown_call(1, user=2, group=3)


def test_chown__changes_ownership_without_following_symlinks(
    tmp_path: Path, mock_os_chown: mock.MagicMock
):
    sh.chown(tmp_path, user=1, group=2, follow_symlinks=False)
    assert mock_os_chown.call_args == chown_call(tmp_path, user=1, group=2, follow_symlinks=False)


def test_chown__changes_ownership_recursively(tmp_path: Path, mock_os_chown: mock.MagicMock):
    test_dir = Dir(
        tmp_path,
        Dir("a", File("1.txt"), File("2.txt"), File("3.txt")),
        Dir(
            "b",
            Dir("c", File("4.txt"), Dir("d", File("5.txt"))),
            File("6.txt"),
            File("7.txt"),
            File("8.txt"),
        ),
        File("9.txt"),
        File("10.txt"),
    )
    test_dir.mkdir()

    sh.chown(test_dir.path, user=1, group=2, recursive=True)

    for path in (test_dir.path, *sh.walk(test_dir.path)):
        assert chown_call(path, user=1, group=2) in mock_os_chown.call_args_list


def test_chown__raises_when_missing_user_and_group():
    with pytest.raises(ValueError):
        sh.chown("path")


def test_chown__raises_when_user_name_invalid():
    with pytest.raises(LookupError):
        sh.chown("path", user=uuid4().hex)


def test_chown__raises_when_group_name_invalid():
    with pytest.raises(LookupError):
        sh.chown("path", group=uuid4().hex)
